"""Compute quantum chemistry using Iowa State's GAMESS executable."""

import copy
import pprint
from decimal import Decimal
from typing import Any, Dict, Optional, Tuple

from qcelemental.models import AtomicInput, AtomicResult, BasisSet, Provenance
from qcelemental.util import safe_version, which

from ...exceptions import InputError, UnknownError
from ...util import execute
from ..model import ProgramHarness
from ..qcvar_identities_resources import build_atomicproperties, build_out
from ..util import error_stamp
from .germinate import muster_modelchem
from .harvester import harvest
from .keywords import format_keywords

pp = pprint.PrettyPrinter(width=120, compact=True, indent=1)


class GAMESSHarness(ProgramHarness):
    """

    Notes
    -----
    Required edits to the ``rungms`` script are as follows::
        set SCR=./                                  # will be managed by QCEngine instead
        set USERSCR=./                              # ditto
        set GMSPATH=/home/psilocaluser/gits/gamess  # full path to installation

    """

    _defaults = {
        "name": "GAMESS",
        "scratch": True,
        "thread_safe": False,
        "thread_parallel": True,
        "node_parallel": True,
        "managed_memory": True,
    }
    version_cache: Dict[str, str] = {}

    class Config(ProgramHarness.Config):
        pass

    @staticmethod
    def found(raise_error: bool = False) -> bool:
        return which(
            "rungms",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Please install via https://www.msg.chem.iastate.edu/GAMESS/GAMESS.html",
        )

    def get_version(self) -> str:
        self.found(raise_error=True)

        which_prog = which("rungms")
        if which_prog not in self.version_cache:
            success, output = execute([which_prog, "v.inp"], {"v.inp": ""})

            if success:
                for line in output["stdout"].splitlines():
                    if "GAMESS VERSION" in line:
                        branch = " ".join(line.strip(" *\t").split()[3:])
                self.version_cache[which_prog] = safe_version(branch)

        return self.version_cache[which_prog]

    def compute(self, input_model: AtomicInput, config: "TaskConfig") -> AtomicResult:
        self.found(raise_error=True)

        job_inputs = self.build_input(input_model, config)
        success, dexe = self.execute(job_inputs)

        if "INPUT HAS AT LEAST ONE SPELLING OR LOGIC MISTAKE" in dexe["stdout"]:
            raise InputError(error_stamp(job_inputs["infiles"]["gamess.inp"], dexe["stdout"], dexe["stderr"]))

        if success:
            dexe["outfiles"]["stdout"] = dexe["stdout"]
            dexe["outfiles"]["stderr"] = dexe["stderr"]
            dexe["outfiles"]["input"] = job_inputs["infiles"]["gamess.inp"]
            return self.parse_output(dexe["outfiles"], input_model)

    def build_input(
        self, input_model: AtomicInput, config: "TaskConfig", template: Optional[str] = None
    ) -> Dict[str, Any]:
        gamessrec = {
            "infiles": {},
            "scratch_directory": config.scratch_directory,
            "scratch_messy": config.scratch_messy,
        }

        opts = copy.deepcopy(input_model.keywords)

        # Handle molecule
        if not all(input_model.molecule.real):
            raise InputError("GAMESS+QCEngine can't handle ghost atoms yet.")

        molcmd, moldata = input_model.molecule.to_string(dtype="gamess", units="Bohr", return_data=True)
        opts.update(moldata["keywords"])

        # Handle calc type and quantum chemical method
        opts.update(muster_modelchem(input_model.model.method, input_model.driver.derivative_int()))

        # Handle basis set
        if isinstance(input_model.model.basis, BasisSet):
            raise InputError("QCSchema BasisSet for model.basis not implemented. Use string basis name.")
        if input_model.model.basis is None:
            raise InputError("None for model.basis is not useable.")

        # * for gamess, usually insufficient b/c either ngauss or ispher needed
        opts["basis__gbasis"] = input_model.model.basis

        # Handle memory
        # * [GiB] --> [M QW]
        # * docs on mwords: "This is given in units of 1,000,000 words (as opposed to 1024*1024 words)"
        # * docs: "the memory required on each processor core for a run using p cores is therefore MEMDDI/p + MWORDS."
        # * int() rounds down
        mwords_total = int(config.memory * (1024**3) / 8e6)

        for mem_frac_replicated in (1, 0.5, 0.1, 0.75):
            mwords, memddi = self._partition(mwords_total, mem_frac_replicated, config.ncores)
            # DEBUG print(f"loop {mwords_total=} {mem_frac_replicated=} {config.ncores=} -> repl: {mwords=} dist: {memddi=} -> percore={memddi/config.ncores + mwords} tot={memddi + config.ncores * mwords}\n")
            trial_opts = copy.deepcopy(opts)
            trial_opts["contrl__exetyp"] = "check"
            trial_opts["system__parall"] = not (config.ncores == 1)
            trial_opts["system__mwords"] = mwords
            trial_opts["system__memddi"] = memddi
            trial_inp = format_keywords(trial_opts) + molcmd
            trial_gamessrec = {
                "infiles": {"trial_gamess.inp": trial_inp},
                "command": [which("rungms"), "trial_gamess"],
                "scratch_messy": False,
                "scratch_directory": config.scratch_directory,
            }
            success, dexe = self.execute(trial_gamessrec)

            # TODO: switch to KnownError and better handle clobbering of user ncores
            # when the "need serial exe" messages show up compared to mem messages isn't clear
            # this would be a lot cleaner if there was a unique or list of memory error strings
            # if (
            #    ("ERROR: ONLY CCTYP=CCSD OR CCTYP=CCSD(T) CAN RUN IN PARALLEL." in dexe["stdout"])
            #    or ("ERROR: ROHF'S CCTYP MUST BE CCSD OR CR-CCL, WITH SERIAL EXECUTION" in dexe["stdout"])
            #    or ("CI PROGRAM CITYP=FSOCI    DOES NOT RUN IN PARALLEL." in dexe["stdout"])
            # ):
            #    print("RESTETTITNG TO 1")
            #    config.ncores = 1
            #    break
            if "INPUT HAS AT LEAST ONE SPELLING OR LOGIC MISTAKE" in dexe["stdout"]:
                raise InputError(error_stamp(trial_inp, dexe["stdout"], dexe["stderr"]))
            elif "EXECUTION OF GAMESS TERMINATED -ABNORMALLY-" in dexe["stdout"]:
                pass
            else:
                opts["system__mwords"] = mwords
                opts["system__memddi"] = memddi
                break

        # TODO: "semget errno=ENOSPC -- check system limit for sysv semaphores." `ipcs -l`, can fix if too many gamess tests run at once by config.ncores = 4

        # ERROR: ROHF'S CCTYP MUST BE CCSD OR CR-CCL, WITH SERIAL EXECUTION
        if opts.get("contrl__cctyp", "").lower() == "ccsd" and opts.get("contrl__scftyp", "").lower() == "rohf":
            config.ncores = 1

        # Handle conversion from schema (flat key/value) keywords into local format
        optcmd = format_keywords(opts)

        gamessrec["infiles"]["gamess.inp"] = optcmd + molcmd
        gamessrec["command"] = [
            which("rungms"),
            "gamess",
            "00",
            str(config.ncores),
        ]  # rungms JOB VERNO NCPUS >& JOB.log &

        return gamessrec

        # Note decr MEMORY=100000 to get
        # ***** ERROR: MEMORY REQUEST EXCEEDS AVAILABLE MEMORY
        # to test gms fail

        # $CONTRL SCFTYP=ROHF MULT=3 RUNTYP=GRADIENT COORD=CART $END
        # $SYSTEM TIMLIM=1 MEMORY=800000 $END
        # $SCF    DIRSCF=.TRUE. $END
        # $BASIS  GBASIS=STO NGAUSS=2 $END
        # $GUESS  GUESS=HUCKEL $END
        # $DATA
        # Methylene...3-B-1 state...ROHF/STO-2G
        # Cnv  2
        #
        # Hydrogen   1.0    0.82884     0.7079   0.0
        # Carbon     6.0
        # Hydrogen   1.0   -0.82884     0.7079   0.0
        # $END

    def execute(self, inputs, extra_outfiles=None, extra_commands=None, scratch_name=None, timeout=None):

        success, dexe = execute(
            inputs["command"],
            inputs["infiles"],
            ["gamess.dat"],
            scratch_messy=inputs["scratch_messy"],
            scratch_directory=inputs["scratch_directory"],
        )
        return success, dexe

    def parse_output(self, outfiles: Dict[str, str], input_model: AtomicInput) -> AtomicResult:

        # Get the stdout from the calculation (required)
        stdout = outfiles.pop("stdout")
        stderr = outfiles.pop("stderr")

        method = input_model.model.method.lower()
        method = method[4:] if method.startswith("gms-") else method

        # gamessmol, if it exists, is dinky, just a clue to geometry of gamess results
        try:
            # July 2021: gamessmol & vector returns now atin/outfile orientation depending on fix_com,orientation=T/F. previously always outfile orientation
            qcvars, gamesshess, gamessgrad, gamessmol, module = harvest(
                input_model.molecule, method, stdout, **outfiles
            )
            # TODO:  "EXECUTION OF GAMESS TERMINATED -ABNORMALLY-" in dexe["stdout"]:

        except Exception:
            raise UnknownError(error_stamp(outfiles["input"], stdout, stderr))

        try:
            if gamessgrad is not None:
                qcvars[f"{method.upper()} TOTAL GRADIENT"] = gamessgrad
                qcvars["CURRENT GRADIENT"] = gamessgrad

            if gamesshess is not None:
                qcvars[f"{method.upper()} TOTAL HESSIAN"] = gamesshess
                qcvars["CURRENT HESSIAN"] = gamesshess

            if input_model.driver.upper() == "PROPERTIES":
                retres = qcvars[f"CURRENT ENERGY"]
            else:
                retres = qcvars[f"CURRENT {input_model.driver.upper()}"]
        except KeyError:
            if "EXETYP=CHECK" in stdout and "EXECUTION OF GAMESS TERMINATED NORMALLY" in stdout:
                # check run that completed normally
                # * on one hand, it's still an error return_result-wise
                # * but on the other hand, often the reason for the job is to get gamessmol, so let it return success=T below
                retres = 0.0
            else:
                raise UnknownError(error_stamp(outfiles["input"], stdout, stderr))

        build_out(qcvars)
        atprop = build_atomicproperties(qcvars)

        provenance = Provenance(creator="GAMESS", version=self.get_version(), routine="rungms").dict()
        if module is not None:
            provenance["module"] = module

        output_data = {
            "schema_version": 1,
            "molecule": gamessmol,  # overwrites with outfile Cartesians in case fix_*=F
            "extras": {**input_model.extras},
            "native_files": {k: v for k, v in outfiles.items() if v is not None},
            "properties": atprop,
            "provenance": provenance,
            "return_result": retres,
            "stderr": stderr,
            "stdout": stdout,
            "success": True,
        }

        # got to even out who needs plump/flat/Decimal/float/ndarray/list
        # * formerly unnp(qcvars, flat=True).items()
        output_data["extras"]["qcvars"] = {
            k.upper(): str(v) if isinstance(v, Decimal) else v for k, v in qcvars.items()
        }

        return AtomicResult(**{**input_model.dict(), **output_data})

    @staticmethod
    def _partition(total: float, fraction_replicated: float, ncores: int) -> Tuple[int, int]:
        """Compute memory keyword values from memory and core parameters.

        Parameters
        ----------
        total
            Total memory per node in mwords.
        fraction_replicated
            Portion on interval (0, 1] to be replicated memory, as opposed to distributed memory.
        ncores
            Number of cores needing replicated memory.

        Returns
        -------
        mwords, memddi
            Return mwords and memddi values for ``total`` memory (already in mwords) partitioned so (0, 1]
        ``fraction_replicated`` is replicated over ``ncores`` and remainder distributed.

        """
        replicated_summed = total * fraction_replicated
        replicated_each = int(max(1, replicated_summed / ncores))
        distributed = int(total - replicated_summed)

        return replicated_each, distributed



from decimal import ROUND_CEILING, ROUND_FLOOR, Decimal

import numpy as np

# This file has considerable history in the quantum chemistry package Psi4


class PreservingDict(dict):
    """Class to store quantum chemical quantities extracted from output
    files. Extends the dictionary object to (1) store key as all-caps
    version of itself and (2) validate value for duplicate values for the
    same key by testing which has more decimal places and whether value
    the same within a plausible rounding error. Allows consistency checks
    when parsing output files without loss of precision.

    works in decimal.Decimal (scalar) and np.ndarray (non-scalar)

    """

    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __setitem__(self, key, value, accept_places=11):
        try:
            key = key.upper()
        except AttributeError:
            raise AttributeError("Keys stored as upper-case strings: %s unsuitable" % (key))

        if isinstance(value, (list, np.ndarray)):
            # non-scalar
            value = np.array(value)

            if key in self.keys() and not "CURRENT" in key:
                if np.allclose(self[key], value, atol=1.0e-5, rtol=1e-3):
                    best_value = value  # no way to choose, really
                else:
                    raise ValueError(
                        """Output file yielded both {} and {} as values for quantity {}.""".format(
                            self[key], value, key
                        )
                    )
                # print("""Resetting array {} to {}""".format(key, best_value))
            else:
                best_value = value
                # print("""Setting   array {} to {}""".format(key, best_value))

        else:
            # scalar
            value = Decimal(value)
            if abs(float(value)) < 1.0e-16:
                value = Decimal("0")

            if key in self.keys() and not "CURRENT" in key:
                # Validate choosing more detailed value for variable
                existing_exp = self[key].as_tuple().exponent  # 0.1111 --> -4
                candidate_exp = value.as_tuple().exponent
                if existing_exp > candidate_exp:  # candidate has more digits
                    places = Decimal(10) ** (existing_exp + 1)  # exp+1 permits slack in rounding
                    best_value = value
                else:  # existing has more digits
                    places = Decimal(10) ** (candidate_exp + 1)
                    best_value = self[key]
                # DEBUG print(f"{existing_exp=} {candidate_exp=} {places=} {self[key]=} {value=}")
                # Validate values are the same
                # places = max(places, Decimal('1E-11'))  # for computed psivars
                places = max(places, Decimal("1E-{}".format(accept_places)))  # for computed psivars
                # print('FLOOR: ', self[key].quantize(places, rounding=ROUND_FLOOR) - value.quantize(places, rounding=ROUND_FLOOR))
                # print('CEIL:  ', self[key].quantize(places, rounding=ROUND_CEILING) - value.quantize(places, rounding=ROUND_CEILING))
                if (
                    self[key]
                    .quantize(places, rounding=ROUND_CEILING)
                    .compare(value.quantize(places, rounding=ROUND_CEILING))
                    != 0
                ) and (
                    self[key]
                    .quantize(places, rounding=ROUND_FLOOR)
                    .compare(value.quantize(places, rounding=ROUND_FLOOR))
                    != 0
                ):
                    raise ValueError(
                        """Output file yielded both %s and %s as values for quantity %s."""
                        % (self[key].to_eng_string(), value.to_eng_string(), key)
                    )
                # print("""Resetting variable {} to {}""".format(key, best_value.to_eng_string()))
            else:
                best_value = value
                # print("""Setting   variable {} to {}""".format(key, best_value.to_eng_string()))

        super(PreservingDict, self).__setitem__(key, best_value)

    def update(self, *args, **kwargs):
        if args:
            if len(args) > 1:
                raise TypeError("update expected at most 1 arguments, " "got %d" % len(args))
            other = dict(args[0])
            for key in other:
                self[key] = other[key]
        for key in kwargs:
            self[key] = kwargs[key]

    def setdefault(self, key, value=None):
        if key not in self:
            self[key] = value
        return self[key]


if __name__ == "__main__":
    c4info = PreservingDict()
    c4info["scf 4.5e0 total energy"] = "-1.e-4"
    c4info["1.3"] = ".4"
    c4info["curl"] = "-437.12345678"
    c4info["curl"] = "-437.12345677"
    c4info["curl"] = "-437.123456"
    c4info["curl"] = "-437.123457"
    c4info["curl"] = "-437.1234444"  # fails
    c4info["curl"] = "-437.123456789"
    c4info["curl"] = "-437.1234567779"  # fails
    print(c4info)



import pytest

from qcengine.programs.terachem_pbs import TeraChemPBSHarness


def test_found_no_tcpb_installed(monkeypatch):
    # Empty sys.path so that no external packages can be found in case tcpb is installed
    with monkeypatch.context() as m:
        m.setattr("sys.path", [])
        harness = TeraChemPBSHarness()
        with pytest.raises(ModuleNotFoundError):
            harness.found(raise_error=True)
        assert harness.found() is False



"""
Calls TeraChem in its "server mode" via a protobuf interface.
"""
import logging
from importlib import import_module
from os import getenv
from typing import TYPE_CHECKING, Any, Dict, Union

from qcelemental.models import AtomicResult, FailedOperation
from qcelemental.util import which_import

from .model import ProgramHarness

if TYPE_CHECKING:
    from qcelemental.models import AtomicInput

    from ..config import TaskConfig

logger = logging.getLogger(__name__)

_pbs_defaults = {
    "name": "terachem_pbs",
    "scratch": False,
    "thread_safe": False,
    "thread_parallel": False,
    "node_parallel": False,
    "managed_memory": True,
}


class TeraChemPBSHarness(ProgramHarness):
    """QCEngine Harness for interfacing with the TeraChem running in Protocol Buffer Server Mode"""

    _defaults = _pbs_defaults
    _tcpb_package: str = "tcpb"
    _tcpb_min_version: str = "0.7.0"
    _tcpb_client: str = "TCProtobufClient"
    _env_vars: Dict[str, Any] = {
        "host": getenv("TERACHEM_PBS_HOST", "127.0.0.1"),
        "port": int(getenv("TERACHEM_PBS_PORT", 11111)),
    }
    _env_vars_external: str = "TERACHEM_PBS_HOST, TERACHEM_PBS_PORT"

    class Config(ProgramHarness.Config):
        pass

    @classmethod
    def found(cls, raise_error: bool = False) -> bool:
        """Whether TeraChemPBS harness is ready for operation.
        Parameters
        ----------
        raise_error: bool
            Passed on to control negative return between False and ModuleNotFoundError raised.
        Returns
        -------
        bool
            If tcpb package is found and server available, returns True.
            If raise_error is False and tcpb package missing and/or server us unavailable, returns False.
            If raise_error is True and tcpb package missing and/or server us unavailable, the error message is raised.
        """
        tcpb_pkg_available = which_import(
            cls._tcpb_package,
            return_bool=True,
            raise_error=raise_error,
            raise_msg=f"TeraChem protobuf client package (tcpb) not found. Please install tcpb>={cls._tcpb_min_version}.",
        )
        if not tcpb_pkg_available:
            return False

        tcpb = import_module(f"{cls._tcpb_package}")

        try:
            with getattr(tcpb, cls._tcpb_client)(**cls._env_vars) as client:
                return client.is_available()
        except tcpb.exceptions.ServerError as e:
            msg = (
                f"Unable to connect to TeraChem server at {cls._env_vars}. If you want to connect to a server elsewhere, consider "
                f"setting environment variables {cls._env_vars_external}."
            )
            logger.error(msg)
            if raise_error:
                raise OSError(msg) from e
        return False

    def get_version(self) -> str:
        """Returns version of TeraChem Protocol Buffer Server"""
        try:
            tcpb = import_module(self._tcpb_package)
        except ModuleNotFoundError:
            return None
        else:
            try:
                return tcpb.__version__
            except AttributeError:
                return None

    def compute(
        self, input_model: "AtomicInput", config: "TaskConfig" = None
    ) -> Union["AtomicResult", "FailedOperation"]:
        """
        Submit AtomicInput to TeraChem running in "server mode"
        """
        self.found(raise_error=True)

        tcpb = import_module(self._tcpb_package)

        with getattr(tcpb, self._tcpb_client)(**self._env_vars) as client:
            return client.compute(input_model)



import numpy as np
import pytest
import qcelemental as qcel
from qcelemental.testing import compare_recursive, compare_values

import qcengine as qcng
from qcengine.testing import using


@using("qcore")
@pytest.mark.parametrize(
    "method, energy, gradient_norm",
    [
        ({"method": "gfn1"}, -5.765990520597323, 0.06294937),
        ({"method": "wb97xd3", "basis": "6-31g"}, -76.3624189841582, 0.04931507),
        ({"method": "hf", "basis": "6-31g"}, -75.98014477585764, 0.08866491),
    ],
)
def test_qcore_methods(method, energy, gradient_norm):

    atomic_input = qcel.models.AtomicInput(
        molecule=qcng.get_molecule("water"),
        model=method,
        driver="gradient",
    )

    atomic_result = qcng.compute(atomic_input, "qcore")

    assert atomic_result.success, atomic_result.error.error_message
    assert compare_values(atomic_result.properties.return_energy, energy)
    assert compare_values(np.linalg.norm(atomic_result.return_result), gradient_norm)
    assert atomic_result.wavefunction is None


@using("qcore")
def test_qcore_wavefunction():

    atomic_input = qcel.models.AtomicInput(
        molecule=qcng.get_molecule("water"),
        model={"method": "wb97xd3", "basis": "6-31g"},
        driver="gradient",
        protocols={"wavefunction": "all"},
    )

    atomic_result = qcng.compute(atomic_input, "qcore")
    assert atomic_result.success, atomic_result.error.error_message
    assert atomic_result.wavefunction is not None
    assert atomic_result.wavefunction.scf_orbitals_a is not None

    assert atomic_result.properties.calcinfo_nalpha == 5
    assert atomic_result.properties.calcinfo_nbasis == 13



"""
Tests the DQM compute dispatch module
"""
import pprint
import re
import tempfile
from pathlib import Path

import pytest
from qcelemental.models import AtomicInput

import qcengine as qcng
from qcengine.testing import has_program, using

_canonical_methods = [
    # needs attn ("adcc", {"method": "adc2", "basis": "6-31G"}, {"n_triplets": 3}),
    ("cfour", {"method": "hf", "basis": "6-31G"}, {}),
    ("dftd3", {"method": "b3lyp-d3"}, {}),
    ("gamess", {"method": "hf", "basis": "n31"}, {"basis__NGAUSS": 6}),
    ("gcp", {"method": "hf3c"}, {}),
    ("mctc-gcp", {"method": "dft/sv"}, {}),
    # needs attn ("molpro", {"method": "hf", "basis": "6-31G"}, {}),
    # needs attn ("mopac", {"method": "PM6"}, {}),
    ("mp2d", {"method": "MP2-DMP2"}, {}),
    # needs attn ("mrchem", {"method": "blyp"}, {"world_prec": 1.0e-3}),
    ("nwchem", {"method": "hf", "basis": "6-31G"}, {}),
    # needs attn ("openmm", {"method": "openff-1.0.0", "basis": "smirnoff"}, {}),
    ("psi4", {"method": "hf", "basis": "6-31G"}, {"gradient_write": True}),
    # needs attn ("qchem", {"method": "hf", "basis": "6-31G"}, {}),
    # needs attn ("qcore", {"method": "pbe", "basis": "6-31G"}, {}),
    # needs attn ("rdkit", {"method": "UFF"}, {}),
    # needs attn ("terachem", {"method": "bad"}, {}),
    # needs attn ("terachem_pbs", {"method": "b3lyp", "basis": "6-31G"}, {}),
    # needs attn ("torchani", {"method": "ANI1x"}, {}),
    # needs attn ("turbomole", {"method": "pbe", "basis": "6-31G"}, {}),
    # needs attn ("xtb", {"method": "GFN2-xTB"}, {}),
]


def _get_molecule(program, method):
    if program in ["openmm", "terachem_pbs"]:
        return qcng.get_molecule("water")
    elif program == "gamess" and method == "ccsd(t)":
        return qcng.get_molecule("water")
    else:
        return qcng.get_molecule("hydrogen")


@pytest.mark.parametrize(
    "memory_trickery",
    [
        pytest.param({}, id="none"),
        pytest.param(
            # native keywords that CONTRADICT config.memory below
            {
                "cfour": {"memory_size": "5000"},
                "gamess": {"system__mwords": "5000"},
                "nwchem": {"memory": "5 gb"},
                "psi4": {},  # no contradictory memory keyword in psi
            },
            id="dsl",
        ),
    ],
)
@pytest.mark.parametrize("program, model, keywords", _canonical_methods)
def test_local_options_memory_gib(program, model, keywords, memory_trickery, request):
    """Ensure memory handling implemented in harness (if applicable).

    For available harnesses, run minimal calc at specific total node memory, both through runtime
      config alone and with clashing (and non-QCEngine-like) keyword spec. Check memory quantity
      shows up in ``TaskConfig``.
    For ``managed-memory``-active harnesses, check that memory registers in output.

    New Harness Instructions
    ------------------------
    * Make sure minimal calc is in _canonical_methods above.
    * If ``managed_memory=True`` in harness, add regex to ``stdout_ref`` below to check that memory
      is specifiable.
    * If this test doesn't work, implement or adjust ``config.memory`` in your harness.

    """
    if not has_program(program):
        pytest.skip(f"Program '{program}' not found.")

    harness = qcng.get_program(program)
    molecule = _get_molecule(program, model["method"])

    addl_keywords = memory_trickery.get(program, memory_trickery)
    use_keywords = {**keywords, **addl_keywords}

    #  <<  Config

    config = qcng.config.get_config(
        hostname="something",
        task_config={
            "ncores": 1,
            "nnodes": 1,
            "memory": 1.555,
        },
    )

    #  <<  Run

    inp = AtomicInput(molecule=molecule, driver="energy", model=model, keywords=use_keywords)
    ret = qcng.compute(inp, program, raise_error=True, task_config=config.dict())
    pprint.pprint(ret.dict(), width=200)
    assert ret.success is True

    #  <<  Reference

    stdout_ref = {  # 1.555 GiB = 208708567 quad-words
        "cfour": "Allocated    1592 MB of main memory",
        "gamess": "208000000 WORDS OF MEMORY AVAILABLE",
        "nwchem": r"total    =  2087085\d\d doubles =   1592.3 Mbytes",  # doubles is quad-words. Mbytes is MiB
        "psi4": "1592 MiB Core",
    }

    #  <<  Test

    assert config.ncores == 1
    assert pytest.approx(config.memory, 0.1) == 1.555

    if harness._defaults["managed_memory"] is True:
        assert re.search(stdout_ref[program], ret.stdout), f"Memory pattern not found: {stdout_ref[program]}"


@pytest.mark.parametrize("program, model, keywords", _canonical_methods)
def test_local_options_scratch(program, model, keywords):
    """Ensure scratch handling implemented in harness (if applicable).

    For available harnesses, run minimal calc at specific scratch directory name (randomly generated
      during test) and skip scratch clean-up. Check scratch settings show up in ``TaskConfig``.
    For ``scratch``-active harnesses, check that an expected file is written to and left behind in
      scratch directory. Check any scratch-related printing in output.

    New Harness Instructions
    ------------------------
    * Make sure minimal calc is in _canonical_methods above.
    * If ``scratch=True`` in harness, add single file (preferrably output) glob to ``scratch_sample``
      below to check that program scratch is directable.
    * If ``scratch=True`` in harness, if scratch directory mentioned in output, add regex to
      ``stdout_ref`` below to check that program scratch is directable. Otherwise, add an
      always-passing regex.
    * If this test doesn't work, implement or adjust ``config.scratch_directory`` and
      ``config.scratch_messy`` in your harness.

    """
    if not has_program(program):
        pytest.skip(f"Program '{program}' not found.")

    harness = qcng.get_program(program)
    molecule = _get_molecule(program, model["method"])

    #  <<  Config

    scratch_directory = tempfile.mkdtemp(suffix="_" + program)

    config = qcng.config.get_config(
        hostname="something",
        task_config={
            "scratch_directory": scratch_directory,
            "scratch_messy": True,
        },
    )

    #  <<  Run

    inp = AtomicInput(molecule=molecule, driver="energy", model=model, keywords=keywords)
    ret = qcng.compute(inp, program, raise_error=True, task_config=config.dict())
    pprint.pprint(ret.dict(), width=200)
    assert ret.success is True

    #  <<  Reference

    stdout_ref = {
        "cfour": "University of Florida",  # freebie
        "dftd3": "Grimme",  # freebie
        "gamess": "IOWA STATE UNIVERSITY",  # freebie
        "gcp": "Grimme",  # freebie
        "mctc-gcp": "Grimme",  # freebie
        "mp2d": "Beran",  # freebie
        "nwchem": "E. Apra",  # freebie
        "psi4": rf"Scratch directory: {scratch_directory}/tmp\w+_psi_scratch/",
    }

    # a scratch file (preferrably output) expected after job if scratch not cleaned up
    scratch_sample = {
        "cfour": "*/NEWFOCK",
        "dftd3": "*/dftd3_geometry.xyz",  # no outfiles
        "gamess": "*/gamess.dat",
        "gcp": "*/gcp_geometry.xyz",  # no outfiles
        "mctc-gcp": "*/gcp_geometry.xyz",  # no outfiles
        "mp2d": "*/mp2d_geometry",  # no outfiles
        "nwchem": "*/nwchem.db",
        "psi4": "*/psi.*.35",
    }

    #  <<  Test

    assert config.scratch_directory.endswith(program)

    if harness._defaults["scratch"] is True:
        sample_file = list(Path(scratch_directory).glob(scratch_sample[program]))
        assert len(sample_file) == 1, f"Scratch sample not found: {scratch_sample[program]} in {scratch_directory}"

        assert re.search(stdout_ref[program], ret.stdout), f"Scratch pattern not found: {stdout_ref[program]}"


@pytest.mark.parametrize("ncores", [1, 3])
@pytest.mark.parametrize("program, model, keywords", _canonical_methods)
def test_local_options_ncores(program, model, keywords, ncores):
    """Ensure multithreading implemented in harness (if applicable) or multithreaded runs don't
    break harness (if inapplicable).

    For available harnesses, run minimal calc with single and multiple cores; check ncores count
      shows up in ``TaskConfig``.
    For ``thread_parallel``-active harnesses, check ncores count registers in output.

    New Harness Instructions
    ------------------------
    * Make sure minimal calc is in _canonical_methods above.
    * If ``thread_parallel=True`` in harness, add regex to ``stdout_ref`` below to check ncores the
      program sees.
    * If this test doesn't work, implement or adjust ``config.ncores`` in your harness.

    """
    if not has_program(program):
        pytest.skip(f"Program '{program}' not found.")

    harness = qcng.get_program(program)
    molecule = _get_molecule(program, model["method"])

    #  <<  Config

    config = qcng.config.get_config(
        hostname="something",
        task_config={
            "ncores": ncores,
            "nnodes": 1,
        },
    )

    #  <<  Run

    inp = AtomicInput(molecule=molecule, driver="energy", model=model, keywords=keywords)
    ret = qcng.compute(inp, program, raise_error=True, task_config=config.dict())
    pprint.pprint(ret.dict(), width=200)
    assert ret.success is True

    #  <<  Reference

    stdout_ref = {
        "cfour": rf"Running with {ncores} threads/proc",
        "gamess": rf"MEMDDI DISTRIBUTED OVER\s+{ncores} PROCESSORS",
        # "gamess": rf"PARALLEL VERSION RUNNING ON\s+{ncores} PROCESSORS IN\s+1 NODES",  # no line for serial
        # nwchem is node_parallel only
        "psi4": rf"Threads:\s+{ncores}",
    }

    #  <<  Test

    assert config.ncores == ncores
    assert config.nnodes == 1

    if harness._defaults["thread_parallel"] is True:
        assert re.search(stdout_ref[program], ret.stdout), f"Thread pattern not found: {stdout_ref[program]}"



"""
Several import utilities
"""

import io
import json
import os
import shutil
import signal
import subprocess
import sys
import tempfile
import time
import traceback
from contextlib import contextmanager
from functools import partial
from pathlib import Path
from threading import Thread
from typing import Any, BinaryIO, Dict, List, Optional, TextIO, Tuple, Union

from pydantic import BaseModel, ValidationError
from qcelemental.models import AtomicResult, FailedOperation, OptimizationResult

from qcengine.config import TaskConfig

from .config import LOGGER, get_provenance_augments
from .exceptions import InputError, QCEngineException

__all__ = ["compute_wrapper", "model_wrapper", "handle_output_metadata", "create_mpi_invocation", "execute"]


def create_mpi_invocation(executable: str, task_config: TaskConfig) -> List[str]:
    """Create the launch command for an MPI-parallel task

    Parameters
    ----------
       executable: str
           Path to executable
       task_config: TaskConfig
           Specification for number of nodes, cores per node, etc.
    """

    # Make the mpirun invocation
    mpirun_str = task_config.mpiexec_command.format(
        nnodes=task_config.nnodes,
        ranks_per_node=task_config.ncores // task_config.cores_per_rank,
        total_ranks=task_config.nnodes * task_config.ncores // task_config.cores_per_rank,
        cores_per_rank=task_config.cores_per_rank,
    )
    command = mpirun_str.split()

    # Add in the desired executable
    command.append(executable)
    return command


def model_wrapper(input_data: Dict[str, Any], model: BaseModel) -> BaseModel:
    """
    Wrap input data in the given model, or return a controlled error
    """

    if isinstance(input_data, dict):
        try:
            input_data = model(**input_data)
        except ValidationError as exc:
            raise InputError(
                f"Error creating '{model.__name__}', data could not be correctly parsed:\n{str(exc)}"
            ) from None
    elif isinstance(input_data, model):
        input_data = input_data.copy()
    else:
        raise InputError("Input type of {} not understood.".format(type(model)))

    # Older QCElemental compat
    try:
        input_data.extras
    except AttributeError:
        input_data = input_data.copy(update={"extras": {}})

    return input_data


@contextmanager
def capture_stdout():
    oldout, olderr = sys.stdout, sys.stderr
    try:
        out = [io.StringIO(), io.StringIO()]
        sys.stdout, sys.stderr = out
        yield out
    finally:
        sys.stdout, sys.stderr = oldout, olderr
        out[0] = out[0].getvalue()
        out[1] = out[1].getvalue()


@contextmanager
def compute_wrapper(capture_output: bool = True, raise_error: bool = False) -> Dict[str, Any]:
    """Wraps compute for timing, output capturing, and raise protection"""

    metadata = {"stdout": None, "stderr": None, "success": True, "retries": 0}

    # Start timer
    comp_time = time.time()

    # Capture stdout/err
    new_stdout = io.StringIO()
    new_stderr = io.StringIO()
    if capture_output:

        old_stdout, sys.stdout = sys.stdout, new_stdout
        old_stderr, sys.stderr = sys.stderr, new_stderr

    try:
        yield metadata

    # Canonical QCEngine, do not show traceback - return error
    except QCEngineException as exc:
        if raise_error:
            raise exc

        metadata["error_type"] = exc.error_type
        metadata["error_message"] = exc.error_message
        metadata["success"] = False

    # Unknown QCEngine exception likely in the Python layer, show traceback
    except Exception as exc:
        if raise_error:
            raise exc

        metadata["error_type"] = "unknown_error"
        metadata["error_message"] = "QCEngine Execution Error:\n" + traceback.format_exc()
        metadata["success"] = False

    # Place data
    metadata["wall_time"] = time.time() - comp_time
    if capture_output:
        sys.stdout = old_stdout
        sys.stderr = old_stderr
        # Pull over values
        metadata["stdout"] = new_stdout.getvalue() or None
        metadata["stderr"] = new_stderr.getvalue() or None


def handle_output_metadata(
    output_data: Union[Dict[str, Any], "AtomicResult", "OptimizationResult", "FailedOperation"],
    metadata: Dict[str, Any],
    raise_error: bool = False,
    return_dict: bool = True,
) -> Union[Dict[str, Any], "AtomicResult", "OptimizationResult", "FailedOperation"]:
    """
    Fuses general metadata and output together.

    Parameters:
        output_data: The original output object to be fused with metadata
        metadata: Metadata produced by the compute_wrapper context manager
        raise_error: Raise an exception if errors exist (True) or return FailedOperation (False)
        return_dict: Return dictionary or object representation of data

    Returns
    -------
    result : AtomicResult, OptimizationResult, FailedOperation, or dict representation of any one.
        Output type depends on return_dict or a dict if an error was generated in model construction
    """

    if isinstance(output_data, dict):
        output_fusion = output_data  # Error handling
    else:
        output_fusion = output_data.dict()

    # Do not override if computer generates
    output_fusion["stdout"] = output_fusion.get("stdout", None) or metadata["stdout"]
    output_fusion["stderr"] = output_fusion.get("stderr", None) or metadata["stderr"]

    if metadata["success"] is not True:
        output_fusion["success"] = False
        output_fusion["error"] = {"error_type": metadata["error_type"], "error_message": metadata["error_message"]}

    # Raise an error if one exists and a user requested a raise
    if raise_error and (output_fusion["success"] is not True):
        msg = "stdout:\n{}".format(output_fusion["stdout"])
        msg += "\nstderr:\n{}".format(output_fusion["stderr"])
        LOGGER.info(msg)
        raise ValueError(output_fusion["error"]["error_message"])

    # Fill out provenance datadata
    provenance_augments = get_provenance_augments()
    provenance_augments["wall_time"] = metadata["wall_time"]
    if "provenance" in output_fusion:
        output_fusion["provenance"].update(provenance_augments)
    else:
        # Add onto the augments with some missing info
        provenance_augments["creator"] = "QCEngine"
        provenance_augments["version"] = provenance_augments["qcengine_version"]
        output_fusion["provenance"] = provenance_augments

    if metadata["retries"] != 0:
        output_fusion["provenance"]["retries"] = metadata["retries"]

    # Make sure pydantic sparsity is upheld
    for val in ["stdout", "stderr"]:
        if output_fusion[val] is None:
            output_fusion.pop(val)

    # We need to return the correct objects; e.g. Results, Procedures
    if output_fusion["success"]:
        # This will only execute if everything went well
        ret = output_data.__class__(**output_fusion)
    else:
        # Should only be reachable on failures
        ret = FailedOperation(
            success=output_fusion.pop("success", False), error=output_fusion.pop("error"), input_data=output_fusion
        )

    if return_dict:
        return json.loads(ret.json())  # Use Pydantic to serialize, then reconstruct as Python dict of Python Primals
    else:
        return ret


def terminate_process(proc: Any, timeout: int = 15) -> None:
    if proc.poll() is None:

        # Sigint (keyboard interupt)
        if sys.platform.startswith("win"):
            proc.send_signal(signal.CTRL_BREAK_EVENT)
        else:
            proc.send_signal(signal.SIGINT)

        try:
            start = time.time()
            while (proc.poll() is None) and (time.time() < (start + timeout)):
                time.sleep(0.02)

        # Flat kill
        finally:
            proc.kill()


@contextmanager
def popen(
    args: List[str],
    append_prefix: bool = False,
    popen_kwargs: Optional[Dict[str, Any]] = None,
    pass_output_forward: bool = False,
) -> Dict[str, Any]:
    """
    Opens a background task

    Code and idea from dask.distributed's testing suite
    https://github.com/dask/distributed

    Parameters
    ----------
        args: List[str]
            Input arguments for the command
        append_prefix: bool
            Whether to prepend the Python path prefix to the command being executed
        popen_kwargs: Dict[str, Any]
            Any keyword arguments to use when launching the process
        pass_output_forward: bool
            Whether to pass the stdout and stderr forward to the system's stdout and stderr
    Returns
    -------
        exe: dict
            Dictionary with the following keys:
            <ul>
                <li>proc: Popen object describing the background task</li>
                <li>stdout: String value of the standard output of the task</li>
                <li>stdeer: String value of the standard error of the task</li>
            </ul>
    """
    args = list(args)
    if popen_kwargs is None:
        popen_kwargs = {}
    else:
        popen_kwargs = popen_kwargs.copy()

    # Bin prefix
    if sys.platform.startswith("win"):
        bin_prefix = os.path.join(sys.prefix, "Scripts")
    else:
        bin_prefix = os.path.join(sys.prefix, "bin")

    # Do we prefix with Python?
    if append_prefix:
        args[0] = os.path.join(bin_prefix, args[0])

    if sys.platform.startswith("win"):
        # Allow using CTRL_C_EVENT / CTRL_BREAK_EVENT
        popen_kwargs["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP

    # Route the standard error and output
    popen_kwargs["stdout"] = subprocess.PIPE
    popen_kwargs["stderr"] = subprocess.PIPE

    # Prepare StringIO objects to store the stdout and stderr
    stdout = io.BytesIO()
    stderr = io.BytesIO()

    # Launch the process
    LOGGER.info("Popen", args, popen_kwargs)

    # Ready the output
    ret = {"proc": subprocess.Popen(args, **popen_kwargs)}

    # Spawn threads that will read from the stderr/stdout
    #  The PIPE uses a buffer with finite capacity. The underlying
    #  process will stall if it is unable to write to the buffer
    #  because the buffer is full. These threads continuously read
    #  from the buffers to ensure that they do not fill.
    #
    def read_from_buffer(buffer: BinaryIO, storage: io.BytesIO, sysio: TextIO):
        for r in iter(partial(buffer.read, 1024), b""):
            storage.write(r)
            if pass_output_forward:
                sysio.write(r.decode())

    stdout_reader = Thread(target=read_from_buffer, args=(ret["proc"].stdout, stdout, sys.stdout))
    stdout_reader.start()
    stderr_reader = Thread(target=read_from_buffer, args=(ret["proc"].stderr, stderr, sys.stderr))
    stderr_reader.start()

    # Yield control back to the main thread
    try:
        yield ret
    except Exception:
        raise

    finally:
        # Executes on an exception or once the context manager closes
        try:
            terminate_process(ret["proc"])
        finally:
            # Wait for the reader threads to finish
            stdout_reader.join()
            stderr_reader.join()

            # Retrieve the standard output for the process
            ret["stdout"] = stdout.getvalue().decode()
            ret["stderr"] = stderr.getvalue().decode()


@contextmanager
def environ_context(config: Optional["TaskConfig"] = None, env: Optional[Dict[str, str]] = None) -> Dict[str, str]:
    """Temporarily set environment variables inside the context manager and
    fully restore previous environment afterwards.

    Parameters
    ----------
    config : Optional[TaskConfig], optional
        Automatically sets MKL/OMP num threads based off the input config.
    env : Optional[Dict[str, str]], optional
        A dictionary of environment variables to update.

    Yields
    ------
    Dict[str, str]
        The updated environment variables.
    """

    temporary_env = {}
    if config:
        temporary_env["OMP_NUM_THREADS"] = str(config.ncores)
        temporary_env["MKL_NUM_THREADS"] = str(config.ncores)

    if env:
        temporary_env.update(env)

    original_env = {key: os.getenv(key) for key in temporary_env}
    os.environ.update(temporary_env)
    try:
        yield temporary_env
    finally:
        for key, value in original_env.items():
            if value is None:
                del os.environ[key]
            else:
                os.environ[key] = value


def execute(
    command: List[str],
    infiles: Optional[Dict[str, str]] = None,
    outfiles: Optional[List[str]] = None,
    *,
    as_binary: Optional[List[str]] = None,
    scratch_name: Optional[str] = None,
    scratch_directory: Optional[str] = None,
    scratch_suffix: Optional[str] = None,
    scratch_messy: bool = False,
    scratch_exist_ok: bool = False,
    blocking_files: Optional[List[str]] = None,
    timeout: Optional[int] = None,
    interupt_after: Optional[int] = None,
    environment: Optional[Dict[str, str]] = None,
    shell: Optional[bool] = False,
    exit_code: Optional[int] = 0,
) -> Tuple[bool, Dict[str, Any]]:
    """
    Runs a process in the background until complete.

    Returns True if exit code <= exit_code (default 0)

    Parameters
    ----------
    command : list of str
    infiles : Dict[str] = str
        Input file names (names, not full paths) and contents.
        to be written in scratch dir. May be {}.
    outfiles : List[str] = None
        Output file names to be collected after execution into
        values. May be {}.
    as_binary : List[str] = None
        Keys of `infiles` or `outfiles` to be treated as bytes.
    scratch_name : str, optional
        Passed to temporary_directory
    scratch_directory : str, optional
        Passed to temporary_directory
    scratch_suffix : str, optional
        Passed to temporary_directory
    scratch_messy : bool, optional
        Passed to temporary_directory
    scratch_exist_ok : bool, optional
        Passed to temporary_directory
    blocking_files : list, optional
        Files which should stop execution if present beforehand.
    timeout : int, optional
        Stop the process after n seconds.
    interupt_after : int, optional
        Interupt the process (not hard kill) after n seconds.
    environment : dict, optional
        The environment to run in
    shell : bool, optional
        Run command through the shell.
    exit_code: int, optional
        The exit code above which the process is considered failure.

    Raises
    ------
    FileExistsError
        If any file in `blocking` is present

    Examples
    --------
    # execute multiple commands in same dir
    >>> success, dexe = qcng.util.execute(['command_1'], infiles, [], scratch_messy=True)
    >>> success, dexe = qcng.util.execute(['command_2'], {}, outfiles, scratch_messy=False, scratch_name=Path(dexe['scratch_directory']).name, scratch_exist_ok=True)

    """

    # Format inputs
    if infiles is None:
        infiles = {}

    if outfiles is None:
        outfiles = []
    outfiles = {k: None for k in outfiles}

    # Check for blocking files
    if blocking_files is not None:
        for fl in blocking_files:
            if os.path.isfile(fl):
                raise FileExistsError("Existing file can interfere with execute operation.", fl)

    # Format popen
    popen_kwargs = {}
    if environment is not None:
        popen_kwargs["env"] = {k: v for k, v in environment.items() if v is not None}

    # Execute
    with temporary_directory(
        child=scratch_name,
        parent=scratch_directory,
        messy=scratch_messy,
        exist_ok=scratch_exist_ok,
        suffix=scratch_suffix,
    ) as scrdir:
        popen_kwargs["cwd"] = scrdir
        popen_kwargs["shell"] = shell
        with disk_files(infiles, outfiles, cwd=scrdir, as_binary=as_binary) as extrafiles:
            with popen(command, popen_kwargs=popen_kwargs) as proc:
                # Wait for the subprocess to complete or the timeout to expire
                if interupt_after is None:
                    proc["proc"].wait(timeout=timeout)
                else:
                    time.sleep(interupt_after)
                    terminate_process(proc["proc"])
            retcode = proc["proc"].poll()
        proc["outfiles"] = extrafiles
    proc["scratch_directory"] = scrdir

    return retcode <= exit_code, proc


@contextmanager
def temporary_directory(
    child: str = None, *, parent: str = None, suffix: str = None, messy: bool = False, exist_ok: bool = False
) -> str:
    """Create and cleanup a quarantined working directory with a parent scratch directory.

    Parameters
    ----------
    child : str, optional
        By default, `None`, quarantine directory generated through
        `tempfile.mdktemp` so guaranteed unique and safe. When specified,
        quarantine directory has exactly `name`.
    parent : str, optional
        Create directory `child` elsewhere than TMP default.
        For TMP default, see https://docs.python.org/3/library/tempfile.html#tempfile.gettempdir
    suffix : str, optional
        Create `child` with identifying label by passing to ``tempfile.mkdtemp``.
        Encouraged use for debugging only.
    messy : bool, optional
        Leave scratch directory and contents on disk after completion.
    exist_ok : bool, optional
        Run commands in a possibly pre-existing directory.

    Yields
    ------
    str
        Full path of scratch directory.

    Raises
    ------
    FileExistsError
        If `child` specified and directory already exists (perhaps from a
        previous `messy=True` run).

    Examples
    --------
    parent            child     suffix   -->  creates
    ------            -----     ------        -------
    None              None      None     -->  /tmp/tmpliyp1i7x/
    None              None      _anharm  -->  /tmp/tmpliyp1i7x_anharm/
    None              myqcjob   None     -->  /tmp/myqcjob/
    /scratch/johndoe  None      None     -->  /scratch/johndoe/tmpliyp1i7x/
    /scratch/johndoe  myqcjob   None     -->  /scratch/johndoe/myqcjob/

    """
    if child is None:
        tmpdir = Path(tempfile.mkdtemp(dir=parent, suffix=suffix))
    else:
        if parent is None:
            parent = Path(tempfile.gettempdir())
        else:
            parent = Path(parent)
        tmpdir = parent / child
        try:
            os.mkdir(tmpdir)
        except FileExistsError:
            if exist_ok:
                pass
            else:
                raise
    try:
        yield tmpdir

    finally:
        if not messy:
            shutil.rmtree(tmpdir)
            LOGGER.info(f"... Removing {tmpdir}")


@contextmanager
def disk_files(
    infiles: Dict[str, Union[str, bytes]],
    outfiles: Dict[str, None],
    *,
    cwd: Optional[str] = None,
    as_binary: Optional[List[str]] = None,
) -> Dict[str, Union[str, bytes]]:
    """Write and collect files.

    Parameters
    ----------
    infiles : Dict[str] = str
        Input file names (names, not full paths) and contents.
        to be written in scratch dir. May be {}.
    outfiles : Dict[str] = None
        Output file names to be collected after execution into
        values. May be {}.
    cwd : str, optional
        Directory to which to write and read files.
    as_binary : List[str] = None
        Keys in `infiles` (`outfiles`) to be written (read) as bytes, not decoded.

    Yields
    ------
    Dict[str] = str
        outfiles with RHS filled in.

    """
    if cwd is None:
        lwd = Path.cwd()
    else:
        lwd = Path(cwd)
    if as_binary is None:
        as_binary = []
    assert set(as_binary) <= (set(infiles) | set(outfiles))

    try:
        for fl, content in infiles.items():
            omode = "wb" if fl in as_binary else "w"
            filename = lwd / fl
            if filename.parent != lwd:
                filename.parent.mkdir(parents=True, exist_ok=True)
            with filename.open(omode) as fp:
                fp.write(content)
                LOGGER.info(f"... Writing ({omode}): {filename}")

        yield outfiles

    finally:
        for fl in outfiles.keys():
            omode = "rb" if fl in as_binary else "r"
            try:
                filename = lwd / fl
                with open(filename, omode) as fp:
                    outfiles[fl] = fp.read()
                    LOGGER.info(f"... Writing ({omode}): {filename}")
            except (OSError, FileNotFoundError):
                if "*" in fl:
                    gfls = {}
                    for gfl in lwd.glob(fl):
                        with open(gfl, omode) as fp:
                            gfls[gfl.name] = fp.read()
                            LOGGER.info(f"... Writing ({omode}): {gfl}")
                    if not gfls:
                        gfls = None
                    outfiles[fl] = gfls
                else:
                    outfiles[fl] = None



"""Support for using QCEngine as an MDI engine.
For details regarding MDI, see https://molssi.github.io/MDI_Library/html/index.html.

"""
from typing import Any, Dict, List, Optional

import numpy as np
import qcelemental as qcel
from qcelemental.util import which_import

from .compute import compute

try:
    from mdi import (
        MDI_CHAR,
        MDI_COMMAND_LENGTH,
        MDI_DOUBLE,
        MDI_INT,
        MDI_MAJOR_VERSION,
        MDI_Accept_Communicator,
        MDI_Init,
        MDI_MPI_get_world_comm,
        MDI_Recv,
        MDI_Recv_Command,
        MDI_Register_Command,
        MDI_Register_Node,
        MDI_Send,
    )

    use_mdi = True
except ImportError:
    use_mdi = False

try:
    from mpi4py import MPI

    use_mpi4py = True
except ImportError:
    use_mpi4py = False


class MDIServer:
    def __init__(
        self,
        mdi_options: str,
        program: str,
        molecule,
        model,
        keywords,
        raise_error: bool = False,
        local_options: Optional[Dict[str, Any]] = None,
    ):
        """Initialize an MDIServer object for communication with MDI

        Parameters
        ----------
        mdi_options: str
            Options used during MDI initialization.
        program : str
            The program to execute the input with.
        molecule
            The initial state of the molecule.
        model
            The simulation model to use.
        keywords
            Program-specific keywords.
        raise_error : bool, optional
            Determines if compute should raise an error or not.
        local_options : Optional[Dict[str, Any]], optional
            A dictionary of local configuration options
        """

        if not use_mdi:
            raise Exception("Trying to run as an MDI engine, but the MDI Library was not found")

        if MDI_MAJOR_VERSION < 1:
            raise Exception("QCEngine requires version 1.0.0 or higher of the MDI Library")

        # Confirm that the MDI library has been located
        which_import("mdi", raise_error=True, raise_msg="Please install via 'conda install pymdi -c conda-forge'")

        # Initialize MDI
        MDI_Init(mdi_options)

        # Input variables
        self.molecule = molecule
        self.model = model
        self.keywords = keywords
        self.program = program
        self.raise_error = raise_error
        self.local_options = local_options

        # The MDI interface does not currently support multiple fragments
        if len(self.molecule.fragments) != 1:
            raise Exception("The MDI interface does not support multiple fragments")

        # Molecule charge and multiplicity
        self.total_charge = self.molecule.molecular_charge
        self.multiplicity = self.molecule.molecular_multiplicity

        # Flag to track whether the latest molecule specification has been validated
        self.molecule_validated = True

        # Output of most recent compute call
        self.compute_return = None

        # MPI variables
        self.mpi_world = None
        self.world_rank = 0

        # Get correct intra-code MPI communicator
        if use_mpi4py:
            self.mpi_world = MDI_MPI_get_world_comm()
            self.world_rank = self.mpi_world.Get_rank()

            # QCEngine does not currently support multiple MPI ranks
            if self.mpi_world.Get_size() != 1:
                MPI.COMM_WORLD.Abort()

        # Flag to stop listening for MDI commands
        self.stop_listening = False

        # Dictionary of all supported MDI commands
        self.commands = {
            "<@": self.send_node,
            "<NATOMS": self.send_natoms,
            "<COORDS": self.send_coords,
            "<ENERGY": self.send_energy,
            "<FORCES": self.send_forces,
            ">COORDS": self.recv_coords,
            "SCF": self.run_energy,
            "<ELEMENTS": self.send_elements,
            ">ELEMENTS": self.recv_elements,
            "<MASSES": self.send_masses,
            ">MASSES": self.recv_masses,
            "<TOTCHARGE": self.send_total_charge,
            ">TOTCHARGE": self.recv_total_charge,
            "<ELEC_MULT": self.send_multiplicity,
            ">ELEC_MULT": self.recv_multiplicity,
            "EXIT": self.stop,
        }

        # Register the @DEFAULT node
        MDI_Register_Node("@DEFAULT")

        # Register all supported commands
        for c in self.commands.keys():
            MDI_Register_Command("@DEFAULT", c)

        # Set the current node
        self.current_node = "@DEFAULT"

        # Accept a communicator to the driver code
        self.comm = MDI_Accept_Communicator()

    def update_molecule(self, key: str, value):
        """Update the molecule

        Parameters
        ----------
        key : str
            Key of the molecular element to update
        value
            Update value
        """
        if key == "molecular_charge" or key == "molecular_multiplicity":
            # In order to validate correctly, the charges and multiplicities must be set simultaneously
            try:
                self.molecule = qcel.models.Molecule(
                    **{
                        **self.molecule.dict(),
                        **{
                            "molecular_charge": self.total_charge,
                            "fragment_charges": [self.total_charge],
                            "molecular_multiplicity": self.multiplicity,
                            "fragment_multiplicities": [self.multiplicity],
                        },
                    }
                )
                self.molecule_validated = True
            except qcel.exceptions.ValidationError:
                # The molecule didn't validate correctly, but a future >TOTCHARGE or >ELEC_MULT command might fix it
                self.molecule_validated = False
        else:
            try:
                self.molecule = qcel.models.Molecule(**{**self.molecule.dict(), **{key: value}})
                self.molecule_validated = True
            except qcel.exceptions.ValidationError:
                if self.molecule_validated:
                    # This update caused the validation error
                    raise Exception("MDI command caused a validation error")

    # Respond to the <@ command
    def send_node(self) -> str:
        """Send the name of the current node through MDI

        Returns
        -------
        node : str
            Name of the current node
        """
        node = self.current_node
        MDI_Send(node, MDI_COMMAND_LENGTH, MDI_CHAR, self.comm)
        return node

    # Respond to the <NATOMS command
    def send_natoms(self) -> int:
        """Send the number of atoms through MDI

        Returns
        -------
        natom : int
            Number of atoms
        """
        natom = len(self.molecule.geometry)
        MDI_Send(natom, 1, MDI_INT, self.comm)
        return natom

    # Respond to the <COORDS command
    def send_coords(self) -> np.ndarray:
        """Send the nuclear coordinates through MDI

        Returns
        -------
        coords : np.ndarray
            Nuclear coordinates
        """
        natom = len(self.molecule.geometry)

        coords = np.reshape(self.molecule.geometry, (3 * natom))
        MDI_Send(coords, 3 * natom, MDI_DOUBLE, self.comm)

        return coords

    # Respond to the >COORDS command
    def recv_coords(self, coords: Optional[np.ndarray] = None) -> None:
        """Receive a set of nuclear coordinates through MDI and assign them to the atoms in the current molecule

        Parameters
        ----------
        coords : np.ndarray, optional
            New nuclear coordinates. If None, receive through MDI.
        """
        natom = len(self.molecule.geometry)
        if coords is None:
            coords = np.zeros(3 * natom)
            MDI_Recv(3 * natom, MDI_DOUBLE, self.comm, buf=coords)
        new_geometry = np.reshape(coords, (-1, 3))
        self.molecule = qcel.models.Molecule(**{**self.molecule.dict(), **{"geometry": new_geometry}})

    # Respond to the <ENERGY command
    def send_energy(self) -> float:
        """Send the total energy through MDI

        Returns
        -------
        energy : float
            Energy of the system
        """
        # Ensure that the molecule currently passes validation
        if not self.molecule_validated:
            raise Exception("MDI attempting to compute energy on an unvalidated molecule")
        self.run_energy()
        energy = self.compute_return.return_result
        MDI_Send(energy, 1, MDI_DOUBLE, self.comm)
        return energy

    # Respond to the <FORCES command
    def send_forces(self) -> np.ndarray:
        """Send the nuclear forces through MDI

        Returns
        -------
        forces : np.ndarray
            Forces on the nuclei
        """
        # Ensure that the molecule currently passes validation
        if not self.molecule_validated:
            raise Exception("MDI attempting to compute gradients on an unvalidated molecule")

        input = qcel.models.AtomicInput(
            molecule=self.molecule, driver="gradient", model=self.model, keywords=self.keywords
        )
        self.compute_return = compute(
            input_data=input, program=self.program, raise_error=self.raise_error, local_options=self.local_options
        )

        forces = self.compute_return.return_result
        MDI_Send(forces, len(forces), MDI_DOUBLE, self.comm)
        return forces

    # Respond to the SCF command
    def run_energy(self) -> None:
        """Run an energy calculation"""
        input = qcel.models.AtomicInput(
            molecule=self.molecule, driver="energy", model=self.model, keywords=self.keywords
        )
        self.compute_return = compute(
            input_data=input, program=self.program, raise_error=self.raise_error, local_options=self.local_options
        )

    # Respond to the <ELEMENTS command
    def send_elements(self):
        """Send the atomic number of each nucleus through MDI

        Returns
        -------
        elements : :obj:`list` of :obj:`int`
            Element of each atom
        """
        natom = len(self.molecule.geometry)
        elements = [qcel.periodictable.to_atomic_number(self.molecule.symbols[iatom]) for iatom in range(natom)]
        MDI_Send(elements, natom, MDI_INT, self.comm)
        return elements

    # Respond to the >ELEMENTS command
    def recv_elements(self, elements: Optional[List[int]] = None):
        """Receive a set of atomic numbers through MDI and assign them to the atoms in the current molecule

        Parameters
        ----------
        elements : :obj:`list` of :obj:`int`, optional
            New element numbers. If None, receive through MDI.
        """
        natom = len(self.molecule.geometry)
        if elements is None:
            elements = MDI_Recv(natom, MDI_INT, self.comm)

        for iatom in range(natom):
            self.molecule.symbols[iatom] = qcel.periodictable.to_symbol(elements[iatom])

        return elements

    # Respond to the <MASSES command
    def send_masses(self) -> np.ndarray:
        """Send the nuclear masses through MDI

        Returns
        -------
        masses : np.ndarray
            Atomic masses
        """
        natom = len(self.molecule.geometry)
        masses = self.molecule.masses
        MDI_Send(masses, natom, MDI_DOUBLE, self.comm)
        return masses

    # Respond to the >MASSES command
    def recv_masses(self, masses: Optional[List[float]] = None) -> None:
        """Receive a set of nuclear masses through MDI and assign them to the atoms in the current molecule

        Parameters
        ----------
        masses : :obj:`list` of :obj:`float`, optional
            New nuclear masses. If None, receive through MDI.
        """
        natom = len(self.molecule.geometry)
        if masses is None:
            masses = MDI_Recv(natom, MDI_DOUBLE, self.comm)
        self.update_molecule("masses", masses)

    # Respond to the <TOTCHARGE command
    def send_total_charge(self) -> float:
        """Send the total system charge through MDI

        Returns
        -------
        charge : float
            Total charge of the system
        """
        charge = self.molecule.molecular_charge
        MDI_Send(charge, 1, MDI_DOUBLE, self.comm)
        return charge

    # Respond to the >TOTCHARGE command
    def recv_total_charge(self, charge: Optional[float] = None) -> None:
        """Receive the total system charge through MDI

        Parameters
        ----------
        charge : float, optional
            New charge of the system. If None, receive through MDI.
        """
        if charge is None:
            charge = MDI_Recv(1, MDI_DOUBLE, self.comm)
        self.total_charge = charge

        # Allow a validation error here, because a future >ELEC_MULT command might resolve it
        try:
            self.update_molecule("molecular_charge", self.total_charge)
        except qcel.exceptions.ValidationError:
            pass

    # Respond to the <ELEC_MULT command
    def send_multiplicity(self) -> int:
        """Send the electronic multiplicity through MDI

        Returns
        -------
        multiplicity : int
            Multiplicity of the system
        """
        multiplicity = self.molecule.molecular_multiplicity
        MDI_Send(multiplicity, 1, MDI_INT, self.comm)
        return multiplicity

    # Respond to the >ELEC_MULT command
    def recv_multiplicity(self, multiplicity: Optional[int] = None) -> None:
        """Receive the electronic multiplicity through MDI

        Parameters
        ----------
        multiplicity : int, optional
            New multiplicity of the system. If None, receive through MDI.
        """
        if multiplicity is None:
            multiplicity = MDI_Recv(1, MDI_INT, self.comm)
        self.multiplicity = multiplicity

        # Allow a validation error here, because a future >TOTCHARGE command might resolve it
        try:
            self.update_molecule("molecular_multiplicity", self.multiplicity)
        except qcel.exceptions.ValidationError:
            pass

    # Respond to the EXIT command
    def stop(self) -> None:
        """Stop listening for MDI commands"""
        self.stop_listening = True

    # Enter server mode, listening for commands from the driver
    def start(self) -> None:
        """Receive commands through MDI and respond to them as defined by the MDI Standard"""

        while not self.stop_listening:
            if self.world_rank == 0:
                command = MDI_Recv_Command(self.comm)
            else:
                command = None
            if use_mpi4py:
                command = self.mpi_world.bcast(command, root=0)
            if self.world_rank == 0:
                print("MDI command received: " + str(command))

            # Search for this command in self.commands
            found_command = False
            for supported_command in self.commands:
                if not found_command and command == supported_command:
                    # Run the function corresponding to this command
                    self.commands[supported_command]()
                    found_command = True
            if not found_command:
                raise Exception("Unrecognized command: " + str(command))



from .hessparse import load_hessian
from .pdict import PreservingDict
from .ao_reordering import (
    reorder_column_ao_indices,
    reorder_row_and_column_ao_indices,
    cca_ao_order_spherical,
    error_stamp,
    get_ao_conversion,
    mill_qcvars,
)



import abc
import logging
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel
from qcelemental.models import AtomicInput, AtomicResult, FailedOperation

from qcengine.exceptions import KnownErrorException
from qcengine.config import TaskConfig

logger = logging.getLogger(__name__)


class ProgramHarness(BaseModel, abc.ABC):

    _defaults: Dict[str, Any] = {}
    name: str
    scratch: bool
    thread_safe: bool
    thread_parallel: bool
    node_parallel: bool
    managed_memory: bool
    extras: Optional[Dict[str, Any]]

    class Config:
        allow_mutation: False
        extra: "forbid"

    def __init__(self, **kwargs):
        super().__init__(**{**self._defaults, **kwargs})

    @abc.abstractmethod
    def compute(self, input_data: AtomicInput, config: TaskConfig) -> Union[AtomicResult, FailedOperation]:
        """Top-level compute method to be implemented for every ProgramHarness

        Note:
            This method behave in any of the following ways:
                1. Return AtomicResult upon successful completion of a calculation
                2. Return FailedOperation object if an operation was unsuccessful or raised an exception. This is most
                    likely to occur if the underlying QC package has a QCSchema API that catches exceptions and
                    returns them as FailedOperation objects to end users.
                3. Raise an exception if a computation failed. The raised exception will be handled by the
                    qcng.compute() method and either raised or packaged as a FailedOperation object.
        """
        pass

    @staticmethod
    @abc.abstractmethod
    def found(raise_error: bool = False) -> bool:
        """
        Checks if the program can be found.

        Parameters
        ----------
        raise_error : bool, optional
            If True, raises an error if the program cannot be found.

        Returns
        -------
        bool
            Returns True if the program was found, False otherwise.
        """

    ## Utility

    def get_version(self) -> str:
        """Finds program, extracts version, returns normalized version string.

        Returns
        -------
        str
            Return a valid, safe python version string.
        """

    ## Computers

    def build_input(
        self, input_model: AtomicInput, config: TaskConfig, template: Optional[str] = None
    ) -> Dict[str, Any]:
        raise ValueError("build_input is not implemented for {}.", self.__class__)

    def execute(
        self,
        inputs: Dict[str, Any],
        extra_outfiles: Optional[List[str]] = None,
        extra_commands: Optional[List[str]] = None,
        scratch_name: Optional[str] = None,
        timeout: Optional[int] = None,
    ) -> Tuple[bool, Dict[str, Any]]:
        raise ValueError("execute is not implemented for {}.", self.__class__)

    def parse_output(self, outfiles: Dict[str, str], input_model: "AtomicInput") -> "AtomicResult":
        raise ValueError("parse_output is not implemented for {}.", self.__class__)


class ErrorCorrectionProgramHarness(ProgramHarness, abc.ABC):
    """Base class for Harnesses that include logic to correct common errors

    Classes which implement this Harness must override the :meth:`_compute` method
    rather than :meth:`compute`. The ``compute`` method from this class will make
    calls to ``_compute`` with different actions as it attempts to correct errors in the
    input files.

    The error corrections are defined by first implementing a :class:`KnownErrorException`
    that contains logic to determine if a certain error occurs and returns an appropriate
    update to the keywords of an AtomicInput.

    Then, modify the ``_compute`` method of your Harness to run the ``detect_error`` method
    of each ``KnownErrorException`` for the Harness after it finishes performing the computation.
    The ``detect_error`` method will raise an exception that is caught by the
    ``ErrorCorrectionProgramHarness`` and used to determine if/how to re-run the computation.
    """

    def _compute(self, input_data: AtomicInput, config: TaskConfig) -> AtomicResult:
        raise NotImplementedError()

    def compute(self, input_data: AtomicInput, config: TaskConfig) -> AtomicResult:
        # Get the error correction configuration
        error_policy = input_data.protocols.error_correction

        # Create a local copy of the input data
        local_input_data = input_data

        # Run the method and, if it fails, assess if the failure is restartable
        observed_errors = {}  # Errors that have been observed previously
        while True:
            try:
                result = self._compute(local_input_data, config)
                break
            except KnownErrorException as e:
                logger.info(f"Caught a {type(e)} error.")

                # Determine whether this specific type of error is allowed
                correction_allowed = error_policy.allows(e.error_name)
                if not correction_allowed:
                    logger.info(f'Error correction for "{e.error_name}" is not allowed')
                    raise e
                logger.info(f'Error correction for "{e.error_name}" is allowed')

                # Check if it has run before
                # TODO (wardlt): Should we allow errors to be run >1 time?
                previously_run = e.error_name in observed_errors
                if previously_run:
                    logger.info(
                        "Error has been observed before and mitigation did not fix the issue. Raising exception"
                    )
                    raise e

                # Generate and apply the updated keywords
                keyword_updates = e.create_keyword_update(local_input_data)
                new_keywords = local_input_data.keywords.copy()
                new_keywords.update(keyword_updates)
                local_input_data = AtomicInput(**local_input_data.dict(exclude={"keywords"}), keywords=new_keywords)

                # Store the error details and mitigations employed
                observed_errors[e.error_name] = {"details": e.details, "keyword_updates": keyword_updates}

        # Add the errors observed and corrected for, if any
        if len(observed_errors) > 0:
            result.extras["observed_errors"] = observed_errors
        return result



import pprint
import re

import pytest
import qcelemental as qcel
from qcelemental.models import AtomicInput

import qcengine as qcng
from qcengine.testing import has_program, using

from .test_canonical_config import _canonical_methods, _get_molecule


@pytest.mark.parametrize("native", ["none", "input", "all"])
@pytest.mark.parametrize("program, model, keywords", _canonical_methods)
def test_protocol_native(program, model, keywords, native):
    """Ensure native_files protocol implemented in harness.

    For harnesses, run minimal gradient calc with different protocol settings; check expected
      native/DSL files show up in ``AtomicResult``.

    New Harness Instructions
    ------------------------
    * Make sure minimal calc is in _canonical_methods above. This uses gradient for more files.
    * Add regex to ``input_ref`` & ``all_ref`` below to check content of input and any other file.
    * If this test doesn't work, implement or adjust ``native_files`` in your harness.

    """
    if not has_program(program):
        pytest.skip(f"Program '{program}' not found.")

    harness = qcng.get_program(program)
    molecule = _get_molecule(program, model["method"])

    #  <<  Config

    protocols = {
        "native_files": native,
    }
    config = qcng.config.get_config(
        hostname="something",
        task_config={
            "ncores": 1,
            "nnodes": 1,
        },
    )

    #  <<  Run

    inp = AtomicInput(molecule=molecule, driver="gradient", model=model, keywords=keywords, protocols=protocols)
    ret = qcng.compute(inp, program, raise_error=True, local_options=config.dict())
    pprint.pprint(ret.dict(), width=200)
    assert ret.success is True

    #  <<  Reference

    input_ref = {
        "cfour": rf"\*CFOUR\(BASIS=6-31G",
        "dftd3": rf"1.000000     1.261000     1.703000",
        "gamess": rf"\$basis gbasis=n31 ngauss=6 \$end",
        "gcp": rf"level HF3C",
        "mctc-gcp": rf"level DFT/SV",
        "mp2d": rf"--TT_a1=0.944 --TT_a2=0.48",
        "nwchem": rf"H library 6-31G",
        "psi4": rf'"driver": "gradient",',
    }

    all_ref = {
        "cfour": ("GRD", rf"1.0+\s+0.0+\s+0.0+\s+0.03"),
        "dftd3": ("dftd3_geometry.xyz", rf"H\s+0.0+\s+0.0+\s+0.34"),
        "gamess": ("gamess.dat", rf"CLOSED SHELL ORBITALS --- GENERATED AT"),
        "gcp": ("gcp_geometry.xyz", rf"H\s+0.0+\s+0.0+\s+0.34"),
        "mctc-gcp": ("gcp_geometry.xyz", rf"H\s+0.0+\s+0.0+\s+0.34"),
        "mp2d": ("mp2d_geometry", rf"H\s+0.0+\s+0.0+\s+0.34"),
        "nwchem": ("nwchem.grad", rf"0.0, 0.0, 0.03"),
        "psi4": ("psi4.grad", rf"1.0+\s+(-?)0.0+\s+(-?)0.0+\s+0.03"),
    }

    #  <<  Test

    if native == "none":
        if qcel.util.parse_version(qcel.__version__) < qcel.util.parse_version("0.25.0"):
            assert ret.native_files is None
        else:
            assert ret.native_files == {}
    elif native == "input":
        assert list(ret.native_files.keys()) == ["input"]

    if ret.native_files:
        assert "stdout" not in ret.native_files, f"Stdout found in native_files -- clean up the harness"
        assert "stderr" not in ret.native_files, f"Stderr found in native_files -- clean up the harness"
    assert "outfiles" not in ret.extras, f"Outfiles found in extras -- clean up the harness"

    if native in ["input", "all"]:
        assert re.search(
            input_ref[program], ret.native_files["input"]
        ), f"Input file pattern not found: {input_ref[program]}"
    if native == "all" and program != "psi4":  # allow psi4 once native_files PR merged
        fl, snip = all_ref[program]
        assert re.search(snip, ret.native_files[fl]), f"Ancillary file pattern not found in {fl}: {snip}"



from typing import Any, Dict

from qcengine.exceptions import InputError


def muster_modelchem(method: str, derint: int) -> Dict[str, Any]:
    """Converts the QC method into GAMESS keywords."""

    method = method.lower()
    opts = {}

    runtyp = {
        0: "energy",
        1: "gradient",
        2: "hessian",
        #'properties': 'prop',
    }[derint]

    opts["contrl__runtyp"] = runtyp

    if method == "gamess":
        pass

    elif method in ["scf", "hf"]:
        pass

        # opts['contrl__mplevl'] = 0
        # opts['contrl__cityp'] = 'none'
        # opts['contrl__cctyp'] = 'none'

    elif method == "mp2":
        opts["contrl__mplevl"] = 2

    elif method == "lccd":
        opts["contrl__cctyp"] = "lccd"

    elif method == "ccd":
        opts["contrl__cctyp"] = "ccd"

    elif method == "ccsd":
        opts["contrl__cctyp"] = "ccsd"

    elif method in ["ccsd+t(ccsd)", "ccsd(t)"]:
        opts["contrl__cctyp"] = "ccsd(t)"

    elif method == "pbe":
        opts["contrl__dfttyp"] = "pbe"

    elif method == "b3lyp":
        opts["contrl__dfttyp"] = "b3lypv1r"

    elif method == "b3lyp5":
        opts["contrl__dfttyp"] = "b3lyp"

    else:
        raise InputError(f"Method not recognized: {method}")

    return opts



from typing import Any, Dict, Union

from qcelemental.models import OptimizationInput, OptimizationResult
from qcelemental.util import safe_version, which_import

from .model import ProcedureHarness


class GeometricProcedure(ProcedureHarness):

    _defaults = {"name": "geomeTRIC", "procedure": "optimization"}

    version_cache: Dict[str, str] = {}

    class Config(ProcedureHarness.Config):
        pass

    def found(self, raise_error: bool = False) -> bool:
        return which_import(
            "geometric",
            return_bool=True,
            raise_error=raise_error,
            raise_msg="Please install via `conda install geometric -c conda-forge`.",
        )

    def get_version(self) -> str:
        self.found(raise_error=True)

        which_prog = which_import("geometric")
        if which_prog not in self.version_cache:
            import geometric

            self.version_cache[which_prog] = safe_version(geometric.__version__)

        return self.version_cache[which_prog]

    def build_input_model(self, data: Union[Dict[str, Any], "OptimizationInput"]) -> "OptimizationInput":
        return self._build_model(data, OptimizationInput)

    def compute(self, input_model: "OptimizationInput", config: "TaskConfig") -> "OptimizationResult":
        try:
            import geometric
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Could not find geomeTRIC in the Python path.")

        input_data = input_model.dict()

        # Temporary patch for geomeTRIC
        input_data["initial_molecule"]["symbols"] = list(input_data["initial_molecule"]["symbols"])

        # Set retries to two if zero while respecting local_config
        local_config = config.dict()
        local_config["retries"] = local_config.get("retries", 2) or 2
        input_data["input_specification"]["extras"]["_qcengine_local_config"] = local_config

        # Run the program
        output_data = geometric.run_json.geometric_run_json(input_data)

        output_data["provenance"] = {
            "creator": "geomeTRIC",
            "routine": "geometric.run_json.geometric_run_json",
            "version": geometric.__version__,
        }

        output_data["schema_name"] = "qcschema_optimization_output"
        output_data["input_specification"]["extras"].pop("_qcengine_local_config", None)
        if output_data["success"]:
            output_data = OptimizationResult(**output_data)

        return output_data



import numpy as np
import pytest
import qcelemental as qcel
from qcelemental.testing import compare_recursive, compare_values

import qcengine as qcng
from qcengine.testing import qcengine_records, using

qchem_info = qcengine_records("qchem")
qchem_logonly_info = qcengine_records("qchem_logonly")

qchem_forgive = [
    "root.molecule.provenance.version",
    "root.molecule.provenance.routine",
    "root.provenance.version",
    "root.provenance.routine",
    "root.provenance.creator",
    "root.extras",
]


@pytest.mark.parametrize("test_case", qchem_info.list_test_cases())
def test_qchem_output_parser(test_case):

    # Get output file data
    data = qchem_info.get_test_data(test_case)
    inp = qcel.models.AtomicInput.parse_raw(data["input.json"])

    outfiles = qcel.util.deserialize(data["outfiles.msgpack"], "msgpack-ext")
    output = qcng.get_program("qchem", check=False).parse_output(outfiles, inp).dict()
    output.pop("provenance", None)

    output_ref = qcel.models.AtomicResult.parse_raw(data["output.json"]).dict()
    output_ref.pop("provenance", None)
    output_ref.pop("extras", None)
    output.pop("extras", None)

    check, message = compare_recursive(output_ref, output, return_message=True)
    assert check, message


@pytest.mark.parametrize("test_case", qchem_info.list_test_cases())
def test_qchem_input_formatter(test_case):

    # Get input file data
    data = qchem_info.get_test_data(test_case)
    inp = qcel.models.AtomicInput.parse_raw(data["input.json"])

    # TODO add actual comparison of generated input file
    input_file = qcng.get_program("qchem", check=False).build_input(inp, qcng.get_config())
    assert input_file.keys() >= {"commands", "infiles"}


@pytest.mark.parametrize("test_case", qchem_info.list_test_cases())
def test_qchem_input_formatter_template(test_case):

    # Get input file data
    data = qchem_info.get_test_data(test_case)
    inp = qcel.models.AtomicInput.parse_raw(data["input.json"])

    # TODO add actual comparison of generated input file
    input_file = qcng.get_program("qchem", check=False).build_input(inp, qcng.get_config(), template="Test template")
    assert input_file.keys() >= {"commands", "infiles"}


@using("qchem")
@pytest.mark.parametrize("test_case", qchem_info.list_test_cases())
def test_qchem_executor(test_case):
    # Get input file data
    data = qchem_info.get_test_data(test_case)
    inp = qcel.models.AtomicInput.parse_raw(data["input.json"])

    # Run qchem
    result = qcng.compute(inp, "qchem")
    assert result.success is True

    # Get output file data
    output_ref = qcel.models.AtomicResult.parse_raw(data["output.json"])

    atol = 1e-6
    assert compare_recursive(output_ref.return_result, result.return_result, atol=atol)


@using("qchem")
def test_qchem_orientation():

    mol = qcel.models.Molecule.from_data(
        """
        He 0.0  0.7  0.7
        He 0.0 -0.7 -0.7
        """
    )

    # Compare with rotation
    inp = {"molecule": mol, "driver": "gradient", "model": {"method": "HF", "basis": "6-31g"}}
    ret = qcng.compute(inp, "qchem", raise_error=True)
    assert compare_values(np.linalg.norm(ret.return_result, axis=0), [0, 0, 0.00791539])

    # Compare without rotations
    mol_noorient = mol.copy(update={"fix_orientation": True})
    inp = {"molecule": mol_noorient, "driver": "gradient", "model": {"method": "HF", "basis": "6-31g"}}
    ret = qcng.compute(inp, "qchem", raise_error=True)

    assert compare_values(np.linalg.norm(ret.return_result, axis=0), [0, 0.00559696541, 0.00559696541])


@pytest.mark.parametrize("test_case", qchem_logonly_info.list_test_cases())
def test_qchem_logfile_parser(test_case):

    # Get output file data
    data = qchem_logonly_info.get_test_data(test_case)
    outfiles = {"dispatch.out": data["qchem.out"]}
    with pytest.warns(Warning):
        output = qcng.get_program("qchem", check=False).parse_logfile(outfiles).dict()
    output["stdout"] = None

    output_ref = qcel.models.AtomicResult.parse_raw(data["output.json"]).dict()
    for key in list(output["provenance"].keys()):
        if key not in output_ref["provenance"]:
            output["provenance"].pop(key)

    # Modify ref to trim down total data as a molecule is now sparse
    output_ref["molecule"] = {k: v for k, v in output_ref["molecule"].items() if k in output["molecule"]}

    check, message = compare_recursive(output_ref, output, return_message=True, forgive=qchem_forgive)
    assert check, message


@pytest.mark.parametrize("test_case", qchem_info.list_test_cases())
def test_qchem_logfile_parser_qcscr(test_case):

    # Get output file data
    data = qchem_info.get_test_data(test_case)
    outfiles = qcel.util.deserialize(data["outfiles.msgpack"], "msgpack-ext")

    with pytest.warns(Warning):
        output = qcng.get_program("qchem", check=False).parse_logfile(outfiles).dict()
    output["stdout"] = None

    output_ref = qcel.models.AtomicResult.parse_raw(data["output.json"]).dict()
    for key in list(output["provenance"].keys()):
        if key not in output_ref["provenance"]:
            output["provenance"].pop(key)

    output_ref["stdout"] = None

    # Modify ref to trim down total data as a molecule is now sparse
    output_ref["molecule"] = {k: v for k, v in output_ref["molecule"].items() if k in output["molecule"]}

    output_ref["model"]["method"] = output_ref["model"]["method"].lower()
    check, message = compare_recursive(output_ref, output, return_message=True, forgive=qchem_forgive)
    assert check, message



import logging
from decimal import Decimal as Dm
from typing import Any, Dict, List

import numpy as np
from qcelemental.models import AtomicResultProperties

from .util import PreservingDict

logger = logging.getLogger(__name__)


def _difference(args):
    minuend, subtrahend = args
    return minuend - subtrahend


def _product(args):
    multiplicand, multiplier = args
    return multiplicand * multiplier


def _spin_component_scaling_wsing(args):
    os_scale, ss_scale, tot_corl, ss_corl, s_corl = args
    return os_scale * (tot_corl - ss_corl - s_corl) + ss_scale * ss_corl + s_corl


def _dispersion_weighting_wsing(args):
    omega, hb_mtd, dd_mtd, s_corl = args
    return omega * hb_mtd + (1.0 - omega) * dd_mtd + s_corl


def omega(args):
    alpha, beta, ratio = args
    return 0.5 * (1.0 + np.tanh(alpha + beta * ratio))


def _linear(args):
    return sum(c * v for c, v in zip(args))


def _solve_in_turn(args: List, coeff: List) -> List[Dict[str, Any]]:
    """Expands a description of a linear equation of variables `args` and weights `coeff` into all rearrangements of the equation."""

    assert len(args) == len(coeff)
    pv0 = []

    for itgt in range(len(args)):
        non_target_args = args[:]
        non_target_args.pop(itgt)

        non_target_coeff = coeff[:]
        non_target_coeff.pop(itgt)
        solve_by = -1 // coeff[itgt]
        non_target_coeff = [solve_by * c for c in non_target_coeff]

        pv0.append(
            {
                "form": args[itgt],
                "func": lambda vv, cc=non_target_coeff: sum(c * v for c, v in zip(vv, cc)),
                "args": non_target_args,
            }
        )

    return pv0


def qcvar_identities() -> List[Dict[str, Any]]:
    """Define QCVariable identity equations (e.g., method total = method correlation + HF)."""

    pv0 = []

    # MP2
    pv0.extend(_solve_in_turn(args=["MP2 TOTAL ENERGY", "HF TOTAL ENERGY", "MP2 CORRELATION ENERGY"], coeff=[-1, 1, 1]))
    pv0.extend(
        _solve_in_turn(
            args=["MP2 DOUBLES ENERGY", "MP2 SAME-SPIN CORRELATION ENERGY", "MP2 OPPOSITE-SPIN CORRELATION ENERGY"],
            coeff=[-1, 1, 1],
        )
    )
    pv0.extend(
        _solve_in_turn(args=["MP2 CORRELATION ENERGY", "MP2 DOUBLES ENERGY", "MP2 SINGLES ENERGY"], coeff=[-1, 1, 1])
    )

    pv0.extend(
        _solve_in_turn(
            args=["CUSTOM SCS-MP2 TOTAL ENERGY", "CUSTOM SCS-MP2 CORRELATION ENERGY", "HF TOTAL ENERGY"],
            coeff=[-1, 1, 1],
        )
    )

    # SCS-MP2
    pv0.append(
        {
            "form": "SCS-MP2 CORRELATION ENERGY",
            "func": _spin_component_scaling_wsing,
            "args": [
                Dm(6) / Dm(5),
                Dm(1) / Dm(3),
                "MP2 CORRELATION ENERGY",
                "MP2 SAME-SPIN CORRELATION ENERGY",
                "MP2 SINGLES ENERGY",
            ],
        }
    )  # yapf: disable
    pv0.append({"form": "SCS-MP2 TOTAL ENERGY", "func": sum, "args": ["HF TOTAL ENERGY", "SCS-MP2 CORRELATION ENERGY"]})

    # SCS(N)-MP2
    pv0.append(
        {
            "form": "SCS(N)-MP2 CORRELATION ENERGY",
            "func": _spin_component_scaling_wsing,
            "args": [
                Dm(0),
                Dm(1.76),
                "MP2 CORRELATION ENERGY",
                "MP2 SAME-SPIN CORRELATION ENERGY",
                "MP2 SINGLES ENERGY",
            ],
        }
    )  # yapf: disable
    pv0.append(
        {"form": "SCS(N)-MP2 TOTAL ENERGY", "func": sum, "args": ["HF TOTAL ENERGY", "SCS(N)-MP2 CORRELATION ENERGY"]}
    )

    # SCS-MP2-VDW
    # * https://doi.org/10.1080/00268970802641242
    pv0.append(
        {
            "form": "SCS-MP2-VDW CORRELATION ENERGY",
            "func": _spin_component_scaling_wsing,
            "args": [
                Dm(1.28),
                Dm(0.50),
                "MP2 CORRELATION ENERGY",
                "MP2 SAME-SPIN CORRELATION ENERGY",
                "MP2 SINGLES ENERGY",
            ],
        }
    )  # yapf: disable
    pv0.append(
        {"form": "SCS-MP2-VDW TOTAL ENERGY", "func": sum, "args": ["HF TOTAL ENERGY", "SCS-MP2-VDW CORRELATION ENERGY"]}
    )

    # DW-MP2
    # only defined at the (IE) reaction level (like SAPT)
    #    dwmp2['DW-MP2 OMEGA'][mt] = \
    #        rxnm_contract_expand(df.xs('HF TOTAL ENERGY', level='qcvar').xs(mtl[0], level='meta')) / \
    #        rxnm_contract_expand(df.xs('MP2 TOTAL ENERGY', level='qcvar').xs(mtl[1], level='meta'))
    #    df_omega = omega([0.15276, 1.89952, df_omega])

    # pv0.append({
    #    'form': 'DW-MP2 CORRELATION ENERGY',
    #    'func': _dispersion_weighting_wsing,
    #    'args': ['DW_MP2 OMEGA', 'MP2 CORRELATION ENERGY', 'SCS-MP2 CORRELATION ENERGY', 'MP2 SINGLES ENERGY']})
    # pv0.append({
    #    'form': 'DW-MP2 TOTAL ENERGY',
    #    'func': sum,
    #    'args': ['HF TOTAL ENERGY', 'DW-MP2 CORRELATION ENERGY']})

    # MP3
    pv0.extend(_solve_in_turn(args=["MP3 TOTAL ENERGY", "HF TOTAL ENERGY", "MP3 CORRELATION ENERGY"], coeff=[-1, 1, 1]))
    pv0.extend(
        _solve_in_turn(
            args=["MP3 DOUBLES ENERGY", "MP3 SAME-SPIN CORRELATION ENERGY", "MP3 OPPOSITE-SPIN CORRELATION ENERGY"],
            coeff=[-1, 1, 1],
        )
    )
    pv0.extend(
        _solve_in_turn(args=["MP3 CORRELATION ENERGY", "MP3 DOUBLES ENERGY", "MP3 SINGLES ENERGY"], coeff=[-1, 1, 1])
    )

    # MPN
    for mpn in range(4, 20):
        pv0.extend(
            _solve_in_turn(
                args=["MP{} TOTAL ENERGY".format(mpn), "HF TOTAL ENERGY", "MP{} CORRELATION ENERGY".format(mpn)],
                coeff=[-1, 1, 1],
            )
        )

    # CCSD
    pv0.extend(
        _solve_in_turn(args=["CCSD TOTAL ENERGY", "HF TOTAL ENERGY", "CCSD CORRELATION ENERGY"], coeff=[-1, 1, 1])
    )
    pv0.extend(
        _solve_in_turn(
            args=["CCSD DOUBLES ENERGY", "CCSD SAME-SPIN CORRELATION ENERGY", "CCSD OPPOSITE-SPIN CORRELATION ENERGY"],
            coeff=[-1, 1, 1],
        )
    )
    pv0.extend(
        _solve_in_turn(args=["CCSD CORRELATION ENERGY", "CCSD DOUBLES ENERGY", "CCSD SINGLES ENERGY"], coeff=[-1, 1, 1])
    )

    # CCSD(T)
    pv0.extend(
        _solve_in_turn(
            args=["CCSD(T) CORRELATION ENERGY", "CCSD CORRELATION ENERGY", "(T) CORRECTION ENERGY"], coeff=[-1, 1, 1]
        )
    )
    pv0.extend(
        _solve_in_turn(args=["CCSD(T) TOTAL ENERGY", "HF TOTAL ENERGY", "CCSD(T) CORRELATION ENERGY"], coeff=[-1, 1, 1])
    )
    pv0.extend(
        _solve_in_turn(
            args=["CCSD(T) CORRELATION ENERGY", "CCSD CORRELATION ENERGY", "(T) CORRECTION ENERGY"], coeff=[-1, 1, 1]
        )
    )  # duplicate of first so that all minimal combinations covered

    # CCSD[T]
    pv0.extend(
        _solve_in_turn(args=["CCSD[T] TOTAL ENERGY", "HF TOTAL ENERGY", "CCSD[T] CORRELATION ENERGY"], coeff=[-1, 1, 1])
    )
    pv0.extend(
        _solve_in_turn(
            args=["CCSD[T] CORRELATION ENERGY", "CCSD CORRELATION ENERGY", "[T] CORRECTION ENERGY"], coeff=[-1, 1, 1]
        )
    )

    # FCI
    pv0.extend(_solve_in_turn(args=["FCI TOTAL ENERGY", "HF TOTAL ENERGY", "FCI CORRELATION ENERGY"], coeff=[-1, 1, 1]))

    # Generics
    pv0.extend(_solve_in_turn(args=["CC TOTAL ENERGY", "HF TOTAL ENERGY", "CC CORRELATION ENERGY"], coeff=[-1, 1, 1]))
    pv0.extend(_solve_in_turn(args=["CI TOTAL ENERGY", "HF TOTAL ENERGY", "CI CORRELATION ENERGY"], coeff=[-1, 1, 1]))

    # DFT

    #   fctl
    #       TODO want B97 here?
    for fctl in [
        "B3LYP",
        "B3LYP5",
        "WPBE",
        "PBE",
        "CAM-B3LYP",
        "B97",
        "WB97X",
        "SVWN",
        "PW91",
        "BLYP",
        "PW86PBE",
        "FT97",
        "BOP",
        "MPWPW",
        "SOGGA11",
        "BP86",
        "B86BPBE",
        "PW6B95",
        "PBE0",
        "SOGGA11-X",
        "MN15",
    ]:
        pv0.extend(
            _solve_in_turn(
                args=["{} TOTAL ENERGY".format(fctl), "{} FUNCTIONAL TOTAL ENERGY".format(fctl)], coeff=[-1, 1]
            )
        )

    #   fctl + D
    for dash in ["-D2", "-D3", "-D3(BJ)", "-D3M", "-D3M(BJ)"]:
        for fctl in ["B3LYP", "B3LYP5", "PBE", "B97", "BLYP", "BP86", "PBE0", "WPBE"]:
            pv0.extend(
                _solve_in_turn(
                    args=[
                        "{}{} TOTAL ENERGY".format(fctl, dash),
                        "{} FUNCTIONAL TOTAL ENERGY".format("B97-D" if fctl == "B97" else fctl),
                        "{}{} DISPERSION CORRECTION ENERGY".format(fctl, dash),
                    ],
                    coeff=[-1, 1, 1],
                )
            )

    #   fctl + DH
    for fctl in ["B2PLYP", "DSD-PBEP86", "PBE0-2", "B2GPPLYP", "PTPSS", "PWPB95", "DSD-BLYP", "PBE0-DH"]:
        pv0.extend(
            _solve_in_turn(
                args=[
                    "{} TOTAL ENERGY".format(fctl),
                    "{} FUNCTIONAL TOTAL ENERGY".format(fctl),
                    "{} DOUBLE-HYBRID CORRECTION ENERGY".format(fctl),
                ],
                coeff=[-1, 1, 1],
            )
        )

    #   fctl + D + DH
    #   no qcvar for fctl + dh, which would be the more restrictive def
    for dash in ["-D2", "-D3", "-D3(BJ)", "-D3M", "-D3M(BJ)"]:
        for fctl in ["B2PLYP"]:
            pv0.extend(
                _solve_in_turn(
                    args=[
                        "{}{} TOTAL ENERGY".format(fctl, dash),
                        "{} TOTAL ENERGY".format(fctl),
                        "{}{} DISPERSION CORRECTION ENERGY".format(fctl, dash),
                    ],
                    coeff=[-1, 1, 1],
                )
            )

    #   misc.
    pv0.extend(
        _solve_in_turn(
            args=[
                "DLDF+D09 TOTAL ENERGY",
                "DLDF+D09 FUNCTIONAL TOTAL ENERGY",
                "DLDF-DAS2009 DISPERSION CORRECTION ENERGY",
            ],
            coeff=[-1, 1, 1],
        )
    )

    pv0.extend(
        _solve_in_turn(
            args=["WB97X-D TOTAL ENERGY", "WB97X FUNCTIONAL TOTAL ENERGY", "WB97-CHG DISPERSION CORRECTION ENERGY"],
            coeff=[-1, 1, 1],
        )
    )

    # misc.
    pv0.extend(
        _solve_in_turn(
            args=["CURRENT ENERGY", "CURRENT REFERENCE ENERGY", "CURRENT CORRELATION ENERGY"], coeff=[-1, 1, 1]
        )
    )

    return pv0


def build_out(rawvars: Dict[str, Any], verbose: int = 1) -> None:
    """Apply standard QC identities to QCVariables `rawvars` to build more (e.g., correlation from total and HF energies).

    Dictionary `qcvar_identities` has keys with names of QCVariables to be created and values with dictionary of two
    keys: `args`, the QCVariables that contribute to the key and `func`, a functional (or lambda) to combine them.
    This function builds that key QCVariables if all the contributors are available in `rawvars`. Updates internally
    so multiple passes not needed.

    Parameters
    ----------
    verbose : int, optional
        Controls print level. Per-var printing with >=2.

    Returns
    -------
    None
        But input dictionary `rawvars` is updated.

    """
    for action in qcvar_identities():
        pvar = action["form"]
        buildline = """building {} {}""".format(pvar, "." * (50 - len(pvar)))

        data_rich_args = []

        for pv in action["args"]:
            if isinstance(pv, str):
                if pv in rawvars:
                    data_rich_args.append(rawvars[pv])
                else:
                    if verbose >= 2:
                        logger.debug("""{}EMPTY, missing {}""".format(buildline, pv))
                    break
            else:
                data_rich_args.append(pv)
        else:
            result = action["func"](data_rich_args)
            # rawvars[pvar] = result
            # with data coming from file --> variable, looks more precise than it is. hack
            rawvars.__setitem__(pvar, result, 6)
            if verbose >= 1:
                logger.debug("""{}SUCCESS""".format(buildline))

            if pvar == "CURRENT CORRELATION ENERGY" and abs(float(rawvars[pvar])) < 1.0e-16:
                rawvars.pop(pvar)


qcvars_to_atomicproperties = {
    # calcinfo
    "N BASIS FUNCTIONS": "calcinfo_nbasis",
    "N MOLECULAR ORBITALS": "calcinfo_nmo",
    "N ATOMS": "calcinfo_natom",
    "N ALPHA ELECTRONS": "calcinfo_nalpha",
    "N BETA ELECTRONS": "calcinfo_nbeta",
    "NUCLEAR REPULSION ENERGY": "nuclear_repulsion_energy",
    # pre keywords
    "ONE-ELECTRON ENERGY": "scf_one_electron_energy",
    "TWO-ELECTRON ENERGY": "scf_two_electron_energy",
    "CURRENT ENERGY": "return_energy",
    "CURRENT GRADIENT": "return_gradient",
    "CURRENT HESSIAN": "return_hessian",
    # HF keywords
    "HF DIPOLE": "scf_dipole_moment",
    "HF QUADRUPOLE": "scf_quadrupole_moment",
    "HF TOTAL ENERGY": "scf_total_energy",
    "HF ITERATIONS": "scf_iterations",
    "HF TOTAL GRADIENT": "scf_total_gradient",
    "HF TOTAL HESSIAN": "scf_total_hessian",
    # MP2 keywords
    "MP2 SAME-SPIN CORRELATION ENERGY": "mp2_same_spin_correlation_energy",
    "MP2 OPPOSITE-SPIN CORRELATION ENERGY": "mp2_opposite_spin_correlation_energy",
    "MP2 SINGLES ENERGY": "mp2_singles_energy",
    "MP2 DOUBLES ENERGY": "mp2_doubles_energy",
    "MP2 CORRELATION ENERGY": "mp2_correlation_energy",
    "MP2 TOTAL ENERGY": "mp2_total_energy",
    "MP2 DIPOLE": "mp2_dipole_moment",
    # CCSD keywords
    "CCSD SAME-SPIN CORRELATION ENERGY": "ccsd_same_spin_correlation_energy",
    "CCSD OPPOSITE-SPIN CORRELATION ENERGY": "ccsd_opposite_spin_correlation_energy",
    "CCSD SINGLES ENERGY": "ccsd_singles_energy",
    "CCSD DOUBLES ENERGY": "ccsd_doubles_energy",
    "CCSD CORRELATION ENERGY": "ccsd_correlation_energy",
    "CCSD TOTAL ENERGY": "ccsd_total_energy",
    "CCSD DIPOLE": "ccsd_dipole_moment",
    "CCSD ITERATIONS": "ccsd_iterations",
    # CCSD(T) keywords
    "CCSD(T) CORRELATION ENERGY": "ccsd_prt_pr_correlation_energy",
    "CCSD(T) TOTAL ENERGY": "ccsd_prt_pr_total_energy",
    "CCSD(T) DIPOLE": "ccsd_prt_pr_dipole_moment",
    # informal keywords
    # scf_vv10_energy
    "DFT XC ENERGY": "scf_xc_energy",
    "DFT DISPERSION ENERGY": "scf_dispersion_correction_energy",
    # multi-definition
    "SCF TOTAL ENERGY": "scf_total_energy",
}


def build_atomicproperties(qcvars: "PreservingDict") -> AtomicResultProperties:
    """For results extracted from QC output in QCDB terminology, translate to QCSchema terminology.

    Parameters
    ----------
    qcvars : PreservingDict
        Dictionary of calculation information in QCDB QCVariable terminology.

    Returns
    -------
    atprop : AtomicResultProperties
        Object of calculation information in QCSchema AtomicResultProperties terminology.

    """
    atprop = {}
    for pv, dpv in qcvars.items():
        if pv in qcvars_to_atomicproperties:
            atprop[qcvars_to_atomicproperties[pv]] = dpv

    return AtomicResultProperties(**atprop)
